package ch.hslu.appe.order.persistency;

import java.util.ArrayList;
import java.util.List;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.hibernate.criterion.Property;
import org.hibernate.criterion.Restrictions;

import ch.hslu.appe.order.model.Bill;
import ch.hslu.appe.order.model.Order;
import ch.hslu.appe.order.model.OrderPosition;
import ch.hslu.appe.persistency.Persistency;

/**
 * Implements persistency for the Order.
 * @author Felix Rohrer <felix.rohrer@stud.hslu.ch>
 */
public class OrderPersistencyImpl extends Persistency implements OrderPersistency {

    /**
     * Constructor.
     */
    public OrderPersistencyImpl() {
        super("OrderData");
    }

    @Override
    public final boolean createOrder(final Order order) {
        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            order.setId((Integer) session.save(order));

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Add log trace entry
            logger.trace("Persisted a new OrderID '" + order.getId() + ", OrderNumber: " + order.getNumber()
                    + " successfully");

            // Flush, to make sure, that the object is saved
            session.flush();

        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to save the order '" + order + "'.", e);

            return false;
        } finally {
            session.close();
        }
        return true;
    }

    @Override
    public final boolean deleteOrder(final int orderId) {
        // Ignore empty invalid id's
        if (orderId <= 0) {
            return false;
        }

        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        if (!deleteOrderPositions(orderId)) {
            return false;
        }

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            // Create a hql query and execute it against the db
            final String hql = "delete from " + Order.class.getName() + " where id= :id";
            session.createQuery(hql).setInteger("id", orderId).executeUpdate();

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Flush, to make sure, that the object is saved
            session.flush();

            // Add log trace entry
            logger.trace("Deleted the order with the id '" + orderId + "' sucessfully");
        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to delete the order with the id '" + orderId + "'.", e);

            return false;
        } finally {
            session.close();
        }
        return true;
    }

    @Override
    public final boolean deleteOrderByNumber(final int orderNumber) {
        final Order order = getOrder(orderNumber);
        if (order == null) {
            logger.error("Failed to delete the order with the orderNumber '" + orderNumber
                    + "'. OrderNumber not found!");
            return false;
        }

        return deleteOrder(order.getId());
    }

    /**
     * Returns a list with all Bills for a specific Ordernumber.
     * @param orderNumber requested ordernumber for the bills
     * @return list of bills of the requested ordernumber
     */
    public final List<Bill> getBillFromOrder(final int orderNumber) {
        // Ignore empty invalid id's
        if (orderNumber <= 0) {
            return null;
        }

        final Session session = sessionFactory.openSession();
        final List<Bill> bills = new ArrayList<Bill>();
        try {

            final Criteria criteria = session.createCriteria(Bill.class);

            // Fetch all bills
            @SuppressWarnings("rawtypes")
            final List fetchedBills = criteria.list();

            // Add bills to the typed generic list
            for (Object bill : fetchedBills) {
                if (((Bill) bill).getOrder().getNumber() == orderNumber) {
                    bills.add((Bill) bill);
                }
            }

            // Add log trace entry
            logger.trace("Fetched " + bills.size() + " bills for the OrderNumber '" + orderNumber + "' sucessfully");

            return bills;
        } finally {
            session.close();
        }
    }

    /**
     * Returns a list with all bills.
     * @return list of all bills
     */
    public final List<Bill> getBills() {
        final Session session = sessionFactory.openSession();
        final ArrayList<Bill> bills = new ArrayList<Bill>();

        try {
            final Criteria criteria = session.createCriteria(Bill.class);

            // Add the filter
            // TODO no filter implemented so far

            // Fetch all bills
            @SuppressWarnings("rawtypes")
            final List fetchedBills = criteria.list();

            // Add order to the typed generic list
            for (Object bill : fetchedBills) {
                bills.add((Bill) bill);
            }

            // Add log trace entry
            logger.trace("Fetched " + bills.size() + " bills sucessfully");
        } finally {
            session.close();
        }
        return bills;
    }

    @Override
    public final int getNextOrderNumber() {
        int currentNumber = 0;
        final Session session = sessionFactory.openSession();

        final Criteria criteria = session.createCriteria(Order.class);
        criteria.addOrder(Property.forName("id").desc());
        criteria.setMaxResults(1);

        final Order order = (Order) criteria.uniqueResult();
        if (order == null) {
            currentNumber = 0;
        } else {
            currentNumber = order.getNumber();
        }

        return currentNumber + 1;
    }

    @Override
    public final Order getOrder(final int orderNumber) {
        // Ignore empty invalid id's
        if (orderNumber <= 0) {
            return null;
        }

        final Session session = sessionFactory.openSession();
        try {

            final Criteria criteria = session.createCriteria(Order.class);

            // Add the criteria (OrderNumber must match)
            criteria.add(Restrictions.eq("number", orderNumber));

            // Get the order by orderNumber
            final Order order = (Order) criteria.uniqueResult();

            // Add log trace entry
            if (order != null) {
                logger.trace("Fetched the order '" + order + "' sucessfully");
            } else {
                logger.warn("Order '" + orderNumber + "' does not exists!");
            }

            return order;
        } finally {
            session.close();
        }
    }

    @Override
    public final List<Order> getOrders() {
        final Session session = sessionFactory.openSession();
        final List<Order> orders = new ArrayList<Order>();

        try {
            final Criteria criteria = session.createCriteria(Order.class);

            // Add the filter
            // TODO no filter implemented so far

            // Fetch all db orders
            @SuppressWarnings("rawtypes")
            final List fetchedOrders = criteria.list();

            // Add order to the typed generic list
            for (Object order : fetchedOrders) {
                orders.add((Order) order);
            }

            // Add log trace entry
            logger.trace("Fetched " + orders.size() + " orders sucessfully");
        } finally {
            session.close();
        }
        return orders;
    }

    @Override
    public final boolean orderExists(final int orderNumber) {
        // Check for invalid orderNumber
        if (orderNumber <= 0) {
            return false;
        }

        final Session session = sessionFactory.openSession();

        try {
            final Query query = session.createQuery("select 1 from Order t where t.number = :id");

            query.setInteger("id", orderNumber);

            // If there's at least one result, return true, otherwise false
            final boolean exists = (query.uniqueResult() != null);

            // Add log trace entry
            logger.trace("Check for order with the ordernumber '" + orderNumber + "' existency: " + exists);

            return exists;
        } finally {
            session.close();
        }
    }

    @Override
    public final boolean updateOrder(final Order order) {
        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        // TODO Update also all positions, bills, etc...

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            session.update(order);

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Flush, to make sure, that the object is saved
            session.flush();

            // Add log trace entry
            logger.trace("Updated the order '" + order + "' sucessfully");

        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to update the order '" + order + "'.", e);
            return false;
        } finally {
            session.close();
        }
        return true;
    }

    /**
     * Adds a Bill to the Order.
     * @param bill new bill to be added
     * @return true if success otherwise false
     */
    @SuppressWarnings("unused")
    private boolean addBill(final Bill bill) {
        Session session = sessionFactory.openSession();
        Transaction transaction = null;

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            bill.setId((Integer) session.save(bill));

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Add log trace entry
            logger.trace("Persisted a new bill '" + bill + "' successfully");

            // Flush, to make sure, that the object is saved
            session.flush();

        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to save the bill '" + bill + "'.", e);

            return false;
        } finally {
            session.close();
        }
        return true;
    }

    /**
     * Adds a Position to the Order.
     * @param orderPosition new Product, OrderPosition to be added
     * @return true if success otherwise false
     */
    @SuppressWarnings("unused")
    private boolean addPosition(final OrderPosition orderPosition) {
        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            orderPosition.setId((Integer) session.save(orderPosition));

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Add log trace entry
            logger.trace("Persisted a new orderPosition '" + orderPosition + "' successfully");

            // Flush, to make sure, that the object is saved
            session.flush();

        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to save the orderPosition '" + orderPosition + "'.", e);

            return false;
        } finally {
            session.close();
        }
        return true;
    }

    /**
     * Deletes a Bill from the Order.
     * @param billId Bill-ID to be deleted
     * @return true if success otherwise false
     */
    @SuppressWarnings("unused")
    private boolean deleteBill(final int billId) {
        // Ignore empty invalid id's
        if (billId <= 0) {
            return false;
        }

        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            // Create a hql query and execute it against the db
            final String hql = "delete from " + Bill.class.getName() + " where id= :id";
            session.createQuery(hql).setInteger("id", billId).executeUpdate();

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Flush, to make sure, that the object is saved
            session.flush();

            // Add log trace entry
            logger.trace("Deleted the bill with the id '" + billId + "' sucessfully");
        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to delete the bill with the id '" + billId + "'.", e);

            return false;
        } finally {
            session.close();
        }
        return true;
    }

    /**
     * Deletes a Order Position.
     * @param orderId OrderID
     * @return true if success otherwise false
     */
    private boolean deleteOrderPositions(final int orderId) {
        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            // Create a hql query and execute it against the db
            final String hql = "delete from " + OrderPosition.class.getName() + " where orderid= :orderid";
            session.createQuery(hql).setInteger("orderid", orderId).executeUpdate();

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Flush, to make sure, that the object is saved
            session.flush();

            // Add log trace entry
            logger.trace("Deleted all order-positions of the order with the id '" + orderId + "' sucessfully");
        } catch (HibernateException e) {

            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();

                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to delete the order-positions of the order with the id '" + orderId + "'.", e);

            return false;
        } finally {
            session.close();
        }

        return true;
    }

    /**
     * Deletes a Product, OrderPosition from this Order.
     * @param orderPositionId OrderPosition-ID to be deleted
     * @return true if success otherwise false
     */
    @SuppressWarnings("unused")
    private boolean deletePosition(final int orderPositionId) {
        // Ignore empty invalid id's
        if (orderPositionId <= 0) {
            return false;
        }

        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            // Create a hql query and execute it against the db
            final String hql = "delete from " + OrderPosition.class.getName() + " where id= :id";
            session.createQuery(hql).setInteger("id", orderPositionId).executeUpdate();

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Flush, to make sure, that the object is saved
            session.flush();

            // Add log trace entry
            logger.trace("Deleted the OrderPosition with the id '" + orderPositionId + "' sucessfully");
        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to delete the OrderPosition with the id '" + orderPositionId + "'.", e);
            return false;
        } finally {
            session.close();
        }
        return true;
    }

    /**
     * Updates a Bill.
     * @param bill Bill with the new informations
     * @return true if success otherwise false
     */
    @SuppressWarnings("unused")
    private boolean updateBill(final Bill bill) {
        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            session.update(bill);

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Flush, to make sure, that the object is saved
            session.flush();

            // Add log trace entry
            logger.trace("Updated the bill '" + bill + "' sucessfully");

        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to update the bill '" + bill + "'.", e);
            return false;
        } finally {
            session.close();
        }
        return true;
    }

    /**
     * Updates a Position.
     * @param orderPosition OrderPosition with the new informations
     * @return true if success otherwise false
     */
    @SuppressWarnings("unused")
    private boolean updatePosition(final OrderPosition orderPosition) {
        final Session session = sessionFactory.openSession();
        Transaction transaction = null;

        try {
            // Begin a transaction
            transaction = getTransaction(session);

            session.update(orderPosition);

            // Commit the transaction
            if (globalTransaction == null) {
                transaction.commit();
            }

            // Flush, to make sure, that the object is saved
            session.flush();

            // Add log trace entry
            logger.trace("Updated the orderPosition '" + orderPosition + "' sucessfully");

        } catch (HibernateException e) {
            // Roll back the transaction
            if (transaction != null && globalTransaction == null) {
                transaction.rollback();
                logger.trace("Roll back transaction due an internal error");
            }

            // Log the error
            logger.error("Failed to update the orderPosition '" + orderPosition + "'.", e);
            return false;
        } finally {
            session.close();
        }
        return true;
    }
}
