package ch.hslu.appe.usermanagement.persistency;

import java.util.ArrayList;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import ch.hslu.appe.usermanagement.exception.CacheOutdatedException;
import ch.hslu.appe.usermanagement.filter.PlaceFilter;
import ch.hslu.appe.usermanagement.model.Place;
import ch.hslu.appe.usermanagement.persistency.PlacePersistencyImpl;

/**
 * Places Persistency Implementation Test.
 */
public class PlacesPersistencyImplTest {

    private PlacePersistencyImpl persistency;

    /**
     * JUnit Setup.
     */
    @Before
    public final void setUp() {

        persistency = new PlacePersistencyImpl();
    }

    /**
     * Test Filtered Locations.
     */
    @Test
    public final void testFilteredLocations() {

        final PlaceFilter filter = new PlaceFilter();

        // There are 93 Locations starting with "Lu"
        filter.setName("Lu%");

        // Expected to find all of them
        Assert.assertEquals(81, persistency.getPlaces(filter).size());
    }

    /**
     * Test Filtered Locations not mathing.
     */
    @Test
    public final void testFilteredLocationsNotMatching() {

        final PlaceFilter filter = new PlaceFilter();

        // There's no canton with the abbreviation "LUZERN"
        filter.setCanton("LUZERN");

        Assert.assertEquals(0, persistency.getPlaces(filter).size());
    }

    /**
     * Test get Places by ID.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testGetPlacesById() throws CacheOutdatedException {

        Assert.assertEquals("Root", persistency.getPlaceById(3249).getName());
    }

    /**
     * Test get Places by non existent ID.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testGetPlacesByNonExistentId() throws CacheOutdatedException {

        // Try to get a place by a wrong id should end up with a null-result
        Assert.assertNull(persistency.getPlaceById(-1));
    }

    /**
     * Test get Places by non existent Zip Code.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testGetPlacesByNonExistentZipCode() throws CacheOutdatedException {

        // Try to get a place by a wrong id should end up with a null-result
        Assert.assertNull(persistency.getPlaceByZipCode(-1));
    }

    /**
     * Test get Place by Zip Code.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testGetPlacesByZipCode() throws CacheOutdatedException {
        Assert.assertEquals("Root", persistency.getPlaceByZipCode(6037).getName());
    }

    /**
     * Test get Place Completeness.
     */
    @Test
    public final void testGetPlacesCompleteness() {

        // Expected to get all 5306 places
        Assert.assertEquals(5306, persistency.getPlaces().size());
    }

    /**
     * Test get Places within Radius.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testGetPlacesWithinRadius() throws CacheOutdatedException {

        // Test this method with a sample of the place Root, LU (Current home of
        // the developer :))
        final Place originPlace = persistency.getPlaceByZipCode(6037);

        // All places within a radius of 5 kilometers
        final ArrayList<Place> placesWithinRadius = persistency.getPlacesWithinRadius(originPlace, 5000);

        Assert.assertEquals(17, placesWithinRadius.size());
    }

    /**
     * Test Place Exists.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testPlaceExists() throws CacheOutdatedException {

        final Place place = new Place();
        place.setId(3249);
        place.setZipCode(6037);

        Assert.assertTrue(persistency.placeExists(place));
    }

    /**
     * Test Place Exists by ZipCode.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testPlaceExistsByZipCode() throws CacheOutdatedException {
        Assert.assertTrue(persistency.placeExists(6037));
    }

    /**
     * Test Place not Exists.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testPlaceNotExists() throws CacheOutdatedException {

        final Place place = new Place();
        place.setId(-1);
        place.setZipCode(-1);

        Assert.assertFalse(persistency.placeExists(place));
    }

    /**
     * Test Place not Exists by Zip Code.
     * @throws CacheOutdatedException Cache expired
     */
    @Test
    public final void testPlaceNotExistsByZipCode() throws CacheOutdatedException {

        Assert.assertFalse(persistency.placeExists(-1));
    }

}
